@@ -5,7 +5,9 @@
 #include "pytype_contacts.h"
 #include "pytype_basics.h"
 #include "structmember.h"
+#include "python_compatibility.h"
 #include <pi-buffer.h>
+#include "python_compatibility.h"
 
 /*******************************************************
  * Create and delloc methods for objects 
@@ -23,6 +25,7 @@ extern PyObject* PyPiContact_New(PyTypeObject *type, PyObject *args, PyObject *k
   self = (PyPiContact *)type->tp_alloc(type, 0);
   new_Contact(&(self->a));
   SetBasicRecordObjectAttributeDefaults((PyObject*) self, pack_Contact);  
+  self->filters = NULL;
 
   return (PyObject*)self;
 }
@@ -30,17 +33,26 @@ extern PyObject* PyPiContact_New(PyTypeObject *type, PyObject *args, PyObject *k
 extern int PyPiContact_Init(PyObject *self, PyObject *args, PyObject *kwds) {
   PyPiContact* fromcontact = NULL;
   PyPiContact* contact = NULL;
+  PyObject* filters = NULL;
   int i;
   int malloc_failed = 0;
   
-  static char *kwlist[] = {"contact", NULL};
-  if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O", kwlist, 
-				   &fromcontact)) {
+  static char *kwlist[] = {"contact","record_field_filters", NULL};
+  if (!PyArg_ParseTupleAndKeywords(args, kwds, "|OO", kwlist,
+				   &fromcontact, &filters)) {
     return -1;
   }
   
   contact = (PyPiContact*)self;
   /* we have to support calling __init__ more than once */
+  if (contact->filters != NULL) {
+    Py_DECREF(contact->filters);
+    contact->filters = NULL;
+  }
+  if (filters != NULL) {
+    contact->filters = filters;
+    Py_INCREF(filters);
+  }
   free_Contact(&(contact->a));
   if (contact->saved_br.size > 0 && contact->saved_br.buf) {	
     free(contact->saved_br.buf);					
@@ -81,17 +93,21 @@ extern int PyPiContact_Init(PyObject *self, PyObject *args, PyObject *kwds) {
     memcpy(&(contact->a), &(fromcontact->a), sizeof(struct Contact));
     
     /* Now do the pointers */
-    
-    if (fromcontact->a.picture) {
-      contact->a.picture = pi_buffer_new((fromcontact->a.picture)->used);
-      if (contact->a.picture == NULL) {
-	PyErr_SetString(PyExc_MemoryError,"Unable to allocate memory for picture");
-	return -1;	    
+
+    for(i=0;i<MAX_CONTACT_BLOBS;i++) {
+      if (fromcontact->a.blob[i] != NULL) {
+	Contact_add_blob(&(contact->a), fromcontact->a.blob[i]);
+	if (!strncmp(contact->a.blob[i]->type, BLOB_TYPE_PICTURE_ID, 4)) {
+	  contact->a.picture = malloc(sizeof(struct ContactPicture));
+	  contact->a.picture->dirty  = 0;
+	  contact->a.picture->length = contact->a.blob[i]->length - 2;
+	  contact->a.picture->data   = contact->a.blob[i]->data + 2;
+	}
+      } else {
+	contact->a.blob[i] = NULL;
       }
-      (contact->a.picture)->used = (fromcontact->a.picture)->used;
-      memcpy((contact->a.picture)->data,(fromcontact->a.picture)->data,(fromcontact->a.picture)->used);
     }
-    
+
     for(i=0;i<NUM_CONTACT_ENTRIES;i++) {
       if (fromcontact->a.entry[i]) {
 	if ((contact->a.entry[i] = malloc(strlen(fromcontact->a.entry[i])+1)) == NULL) {
@@ -130,14 +146,57 @@ static PyObject * PyPiContact_Allocate(PyTypeObject *type, int nitems) {
   }
 }
 
+/**
+ * Wrap an existing Contact in a jppy contact. This makes
+ * a copy of a into a newly allocated struct inside the python object.
+ *
+ * @param a the palm object
+ * @param rt
+ * @param unique_id
+ * @param attrib
+ * @param size
+ * @param buf
+ * @return a newly created python event object that represents the data in a
+ */
 extern PyObject* PyPiContact_Wrap(struct Contact *a, PCRecType rt, 
 				 unsigned int unique_id, unsigned char attrib,
-				 int size, void* buf) {
+				 int size, void* buf, PyObject *filters) {
   PyPiContact* contact;
   int i;
   int malloc_failed = 0;
-  
-  contact = (PyPiContact*)PyPiContact_New(&ContactType,NULL,NULL);
+
+  PyObject *python_mod, *python_mdict, *contact_class, *python_args, *python_kw;
+  python_mod = PyImport_Import(PyString_FromString("jppy.jpilot.modern"));
+  if (python_mod == NULL) {
+    PyErr_Print();
+    return NULL;
+  }
+  python_mdict = PyModule_GetDict(python_mod);
+  if (python_mdict == NULL) {
+    PyErr_Print();
+    Py_DECREF(python_mod);
+    return NULL;
+  }  
+  Py_INCREF(python_mdict);
+  Py_DECREF(python_mod);
+  contact_class = PyDict_GetItemString(python_mdict, "Contact"); /* borrowed reference */
+  if (contact_class == NULL) {
+    PyErr_Print();
+    Py_DECREF(python_mdict);
+    return NULL;
+  }
+  Py_INCREF(contact_class);
+  python_args = Py_BuildValue("()");
+  python_kw = Py_BuildValue("{s:O}","record_field_filters",filters);
+  contact = (PyPiContact*) PyObject_Call(contact_class, python_args, python_kw);
+  Py_DECREF(contact_class);
+  Py_DECREF(python_args);
+  Py_DECREF(python_kw);
+  if (contact == NULL) {
+    PyErr_Print();
+    return NULL;
+  }
+  Py_INCREF(contact);
 
   /* copy birthday stuff, phoneLabels and showPhone */
   memcpy(&(contact->a), a, sizeof(struct Contact));
@@ -145,17 +204,20 @@ extern PyObject* PyPiContact_Wrap(struct Contact *a, PCRecType rt,
   /* set saved_br stuff, and rt and unique_id, and attrib derived details for the current contact */
   SetSavedBrAndRTandUniqueIDandAttribs(rt, unique_id, attrib, size, buf, (PyObject *)contact);
 
-  /* Now do the pointers */
-  if (a->picture) {
-    contact->a.picture = pi_buffer_new((a->picture)->used);
-    if (contact->a.picture == NULL) {
-      PyErr_SetString(PyExc_MemoryError,"Unable to allocate memory for picture");
-      return NULL;	    
+  for(i=0;i<MAX_CONTACT_BLOBS;i++) {
+    if (a->blob[i] != NULL) {
+      Contact_add_blob(&(contact->a), a->blob[i]);
+      if (!strncmp(contact->a.blob[i]->type, BLOB_TYPE_PICTURE_ID, 4)) {
+	contact->a.picture = malloc(sizeof(struct ContactPicture));
+	contact->a.picture->dirty  = 0;
+	contact->a.picture->length = contact->a.blob[i]->length - 2;
+	contact->a.picture->data   = contact->a.blob[i]->data + 2;
+      }
+    } else {
+      contact->a.blob[i] = NULL;
     }
-    (contact->a.picture)->used = (a->picture)->used;
-    memcpy((contact->a.picture)->data,(a->picture)->data,(a->picture)->used);
   }
-
+  
   for(i=0;i<NUM_CONTACT_ENTRIES;i++) {
     if (a->entry[i]) {
       if ((contact->a.entry[i] = malloc(strlen(a->entry[i])+1)) == NULL) {
@@ -170,8 +232,8 @@ extern PyObject* PyPiContact_Wrap(struct Contact *a, PCRecType rt,
 
   if (malloc_failed) { /* Oh well, undo the mallocs that worked */
     for(i=0;i<NUM_CONTACT_ENTRIES;i++) {
-      if (a->entry[i]) {
-	free(a->entry[i]);
+      if (contact->a.entry[i]) {
+	free(contact->a.entry[i]);
       }
     }
     PyErr_SetString(PyExc_MemoryError,"Unable to allocate memory for contact entires");
@@ -183,6 +245,10 @@ extern PyObject* PyPiContact_Wrap(struct Contact *a, PCRecType rt,
 
 static void PyPiContact_Dealloc(PyPiContact* self) {
   free_Contact(&(self->a));
+  if (self->filters != NULL) {
+    Py_DECREF(self->filters);
+    self->filters = NULL;
+  }
   if (self->saved_br.size > 0 && self->saved_br.buf) {
     free(self->saved_br.buf);
   }
@@ -291,9 +357,11 @@ static PyObject* PyPiContact_keys(PyObject* self) {
     PyList_Append(list, value);
     Py_DECREF(value);
   }
+  PyPi_extend_keys_from_filters((PyPiBase*)self, list);
   return list;
 }
 
+/* forward declaration */
 PyObject *PyPiContact_GetItem(PyPiContact* self,  PyObject* key);
 
 static PyObject* PyPiContact_values(PyObject* self) {
@@ -434,7 +502,7 @@ static PyObject* PyPiContact_Generate_VCard(PyPiContact* self, PyObject *args, P
 
   safeAddPropValue(vcard,VCTitleProp,self->a.entry[contTitle]);
 
-  for (i=0;i<NUM_CONTACT_PHONES;i++) {
+  for (i=0;i<7;i++) {
     if ((self->a.entry[contPhone1 + i]) && (strcmp("",self->a.entry[contPhone1 + i]))) {
       switch(self->a.phoneLabel[i]) {
       case 0:
@@ -526,11 +594,11 @@ static PyObject* PyPiContact_Generate_VCard(PyPiContact* self, PyObject *args, P
 
     safeAddPropValue(vcard,VCCommentProp,self->a.entry[contNote]);
 
-    if (self->a.pictype == cpic_jpeg) {
+    if (self->a.picture != NULL) {
       prop = addProp(vcard,VCPhotoProp);
       setValueWithSize(prop, 
 		       (self->a.picture)->data, 
-		       (self->a.picture)->used);
+		       (self->a.picture)->length);
       addPropValue(prop,"TYPE",VCJPEGProp);
       addPropValue(prop,VCEncodingProp,VCBase64Prop);
     }
@@ -557,7 +625,7 @@ static PyObject* PyPiContact_Generate_VCard(PyPiContact* self, PyObject *args, P
     PYTYPE_CONTACTS_ADD_ADR_TO_VCARD(2);
     PYTYPE_CONTACTS_ADD_ADR_TO_VCARD(3);
 
-    for (i=0;i<NUM_CONTACT_CUSTOMS;i++) {  
+    for (i=0;i<9;i++) {  
       if (self->a.entry[contCustom1 + i]){
 	snprintf(strbuf,20,"X-Palm-Custom%d",i + 1);
 	safeAddPropValue(vcard,strbuf,self->a.entry[contCustom1 + i]);
@@ -576,7 +644,7 @@ static PyMethodDef PyPiContact_Methods[] = {
   { "keys", (PyCFunction)PyPiContact_keys, METH_NOARGS, "Return a list of available keys"},
   { "items",(PyCFunction)PyPiContact_items, METH_NOARGS, "Return a list of available items"},
   { "values",(PyCFunction)PyPiContact_values, METH_NOARGS, "Return a list of available items"},
-  {NULL,NULL}
+  {NULL,NULL,0,NULL}
 };
 
 
@@ -591,13 +659,20 @@ static PyGetSetDef PyPiContact_Getseters[] = {
 };
 
 /**** mapping interface ****/
-int PyPiContact_Len(self) {
-  return 5;
+int PyPiContact_Len(PyObject* self) {
+  int len=0;
+  // quite expensive way to do it, but we need to get filters too
+  PyObject *keys = PyPiContact_keys(self); 
+  len = PySequence_Size(keys);
+  Py_DECREF(keys);
+  return len;
 }
 
 
 PyObject *PyPiContact_GetItem(PyPiContact* self,  PyObject* key) {
   char *keystring;
+  PyObject *result;
+
   PyObject* py_list;
   int n, i;
 
@@ -606,6 +681,11 @@ PyObject *PyPiContact_GetItem(PyPiContact* self,  PyObject* key) {
     return Py_None;  
   }
 
+  if ((result = PyPi_GetItem_from_filters((PyPiBase *)self, key)) != NULL)
+    return result;
+  else if (PyErr_Occurred() != NULL)
+    return NULL;
+
   Py_INCREF(key);
   keystring = PyString_AsString(key);
 
@@ -679,9 +759,9 @@ PyObject *PyPiContact_GetItem(PyPiContact* self,  PyObject* key) {
   GET_INT_ATTR(keystring,"typeaddr3",a.addressLabel[2]);
 
   if (strcasecmp(keystring,"picture") == 0) {
-    if ((self->a.pictype == cpic_jpeg) && (self->a.picture)) {
+    if (self->a.picture != NULL) {
       PyObject *value;
-      value = PyString_FromStringAndSize((char *)((self->a.picture)->data), (self->a.picture)->used);
+      value = PyString_FromStringAndSize((char *)((self->a.picture)->data), (self->a.picture)->length);
       Py_DECREF(key);
       return value;
     } else {
@@ -692,7 +772,7 @@ PyObject *PyPiContact_GetItem(PyPiContact* self,  PyObject* key) {
   }
 
   if (strcasecmp(keystring,"phones") == 0) {
-    py_list = PyList_New(NUM_CONTACT_PHONES);
+    py_list = PyList_New(7);
     PyList_SET_ITEM(py_list, 0, PyPiContact_GetItem(self, PyString_FromString("phone1")));
     PyList_SET_ITEM(py_list, 1, PyPiContact_GetItem(self, PyString_FromString("phone2")));
     PyList_SET_ITEM(py_list, 2, PyPiContact_GetItem(self, PyString_FromString("phone3")));
@@ -705,7 +785,7 @@ PyObject *PyPiContact_GetItem(PyPiContact* self,  PyObject* key) {
   }  
 
   if (strcasecmp(keystring,"phones_with_labels") == 0) {
-    py_list = PyList_New(NUM_CONTACT_PHONES);
+    py_list = PyList_New(7);
     PyList_SET_ITEM(py_list, 0, Py_BuildValue("(Oi)",PyPiContact_GetItem(self, PyString_FromString("phone1")),
 					      self->a.phoneLabel[0]));
     PyList_SET_ITEM(py_list, 1, Py_BuildValue("(Oi)",PyPiContact_GetItem(self, PyString_FromString("phone2")),
@@ -731,7 +811,7 @@ PyObject *PyPiContact_GetItem(PyPiContact* self,  PyObject* key) {
     n = 7;
   }
   if (n > 0) {
-    for(i=0;i<NUM_CONTACT_PHONES;i++) {
+    for(i=0;i<7;i++) {
       if (n == self->a.phoneLabel[i]) {
 	if (self->a.entry[contPhone1 + i] != NULL) {
 	  Py_DECREF(key);
@@ -753,12 +833,18 @@ int PyPiContact_SetItem(PyPiContact* self, PyObject* key, PyObject* value) {
   char buf[255];
   char *keystring;
   PyObject *pytup, *pyint, *pystr;
+  struct ContactPicture picture;
   int n;
 
   if (!PyString_Check(key)) {
     PyErr_SetString(PyExc_TypeError,"key must be a String");
     return -1;
   }
+
+  if (PyPi_SetItem_from_filters((PyPiBase *)self, key, value) > 0)
+    return 0;
+  else if (PyErr_Occurred() != NULL)
+    return -1;
   
   Py_INCREF(key);
   keystring = PyString_AsString(key);
@@ -839,36 +925,49 @@ int PyPiContact_SetItem(PyPiContact* self, PyObject* key, PyObject* value) {
   SET_BOUNDED_INT_ATTR(keystring,"showphone",a.showPhone,value,0,4,buf,255);
 
   if (strcasecmp(keystring,"picture") == 0) {
-    if (value == Py_None) {
-      if (self->a.picture) {
-	pi_buffer_free(self->a.picture);
+
+    // free the old one (if any)
+    if (self->a.picture) {
+      for(n=0;n<MAX_CONTACT_BLOBS;n++) {
+	// free all the picture blobs
+	if (self->a.blob[n] != NULL) {
+	  if (!strncmp(self->a.blob[n]->type, BLOB_TYPE_PICTURE_ID, 4)) {
+	    free(self->a.blob[n]->data);
+	    free(self->a.blob[n]);
+	    self->a.blob[n] = NULL;
+	  }
+	}
       }
+      free(self->a.picture);
       self->a.picture = NULL;
-      self->a.pictype = cpic_none;
+    }
+
+    if (value == Py_None) {
+      // if we're removing the image, we're all done
+      self->unsaved_changes = 1;
       Py_DECREF(key);
       return 0;
     }
+    // otherwise we need to put in the new image
     if (!PyString_Check(value)) {
       PyErr_SetString(PyExc_TypeError,"picture must be set to a string (jpeg data)");
       Py_DECREF(key);
       return -1;
     }
-    int length;
-    char *buffer;
-    pi_buffer_t *pibuffer;
-    PyString_AsStringAndSize(value, &buffer, &length);
-    pibuffer = pi_buffer_new(length);
-    if (pibuffer == NULL) {
-      PyErr_SetString(PyExc_MemoryError,"Unable to allocate memory for picture");
-      return -1;	    
+    PyString_AsStringAndSize(value, (char **)(&picture.data), &picture.length);
+    Contact_add_picture(&(self->a), &picture);
+    // find the picture data again :-/
+    for(n=0;n<MAX_CONTACT_BLOBS;n++) {
+      if (self->a.blob[n] != NULL) {
+	if (!strncmp(self->a.blob[n]->type, BLOB_TYPE_PICTURE_ID, 4)) {
+	  self->a.picture = malloc(sizeof(struct ContactPicture));
+	  self->a.picture->dirty   = 0;
+	  self->a.picture->length = self->a.blob[n]->length - 2;
+	  self->a.picture->data   = self->a.blob[n]->data + 2;
+	  break;
+	}    
+      }
     }
-    if (self->a.picture) {
-      pi_buffer_free(self->a.picture);
-    }
-    self->a.picture = pibuffer;
-    (self->a.picture)->used = length;
-    memcpy((self->a.picture)->data,buffer,length);
-    self->a.pictype = cpic_jpeg;
     self->unsaved_changes = 1;
     Py_DECREF(key);
     return 0;
@@ -880,13 +979,13 @@ int PyPiContact_SetItem(PyPiContact* self, PyObject* key, PyObject* value) {
       Py_DECREF(key);
       return -1;
     }
-    if (PyList_GET_SIZE(value) != NUM_CONTACT_PHONES) {
+    if (PyList_GET_SIZE(value) != 7) {
       PyErr_SetString(PyExc_IndexError,"phone_with_labels must have 7 elements");
       Py_DECREF(key);
       return -1;
     }
 
-    for (n=0;n<NUM_CONTACT_PHONES;n++) {
+    for (n=0;n<7;n++) {
       pytup = PyList_GET_ITEM(value,n);
       if (!PyTuple_Check(pytup)) {
 	PyErr_SetString(PyExc_TypeError,"phone_with_labels must be set to a list containing tuples");
@@ -931,7 +1030,7 @@ int PyPiContact_SetItem(PyPiContact* self, PyObject* key, PyObject* value) {
 }
 
 static PyMappingMethods PyPiContact_Mapping = {
-  (inquiry)PyPiContact_Len,
+  (lenfunc)PyPiContact_Len,
   (binaryfunc)PyPiContact_GetItem, 
   (objobjargproc)PyPiContact_SetItem,
 };
